# Copyright (c) Alibaba, Inc. and its affiliates.
locale_mapping = {
    'modify_system': {
        'en': '🛠️ Set system and clear history',
        'zh': '🛠️ 设置system并清空历史'
    },
    'clear_history': {
        'en': '🧹 Clear history',
        'zh': '🧹 清空历史'
    },
    'submit': {
        'en': '🚀 Send',
        'zh': '🚀 发送'
    },
    'regenerate': {
        'en': '🤔️ Regenerate',
        'zh': '🤔️ 重试'
    },
    'upload': {
        'en': '📁 Upload',
        'zh': '📁 上传'
    }
}
