# Copyright (c) Alibaba, Inc. and its affiliates.

from .config import convert_hf_config
from .dequantizer import Fp8Dequantizer, MxFp4Dequantizer
from .io_utils import LazyTensor, SafetensorLazyLoader, StreamingSafetensorSaver
from .patcher import patch_load_base_checkpoint, patch_merge_fn, patch_torch_dist_shard
from .utils import (MegatronTrainerState, adapter_state_dict_context, copy_original_module_weight, forward_step_helper,
                    get_local_layer_specs, get_padding_to, prepare_mcore_model, tuners_sharded_state_dict)
